import datetime
from typing import Any, cast

import pytest

from mycli.packages import key_binding_utils


class FakeSQLExecute:
    def __init__(self, now_value: datetime.datetime) -> None:
        self.now_value = now_value

    def now(self) -> datetime.datetime:
        return self.now_value


class FakePromptSession:
    def __init__(self, responses: list[object]) -> None:
        self.responses = list(responses)
        self.prompt_calls: list[dict[str, Any]] = []

    def prompt(self, *, default: str, inputhook: Any, message: Any) -> str:
        self.prompt_calls.append({
            'default': default,
            'inputhook': inputhook,
            'message': message,
        })
        response = self.responses.pop(0)
        if isinstance(response, BaseException):
            raise response
        return cast(str, response)


class FakeMyCli:
    def __init__(
        self,
        *,
        prompt_session: FakePromptSession | None = None,
        last_query: str = 'last query',
    ) -> None:
        self.prompt_session = prompt_session
        self.last_query = last_query
        self.toolbar_error_message: str | None = None

    def get_last_query(self) -> str:
        return self.last_query


def test_server_date_returns_quoted_and_unquoted_values() -> None:
    sqlexecute = FakeSQLExecute(datetime.datetime(2026, 4, 3, 14, 5, 6))

    assert key_binding_utils.server_date(cast(Any, sqlexecute)) == '2026-04-03'
    assert key_binding_utils.server_date(cast(Any, sqlexecute), quoted=True) == "'2026-04-03'"


def test_server_datetime_returns_quoted_and_unquoted_values() -> None:
    sqlexecute = FakeSQLExecute(datetime.datetime(2026, 4, 3, 14, 5, 6))

    assert key_binding_utils.server_datetime(cast(Any, sqlexecute)) == '2026-04-03 14:05:06'
    assert key_binding_utils.server_datetime(cast(Any, sqlexecute), quoted=True) == "'2026-04-03 14:05:06'"


def test_prettify_statement():
    statement = 'SELECT 1'
    mycli = FakeMyCli()
    pretty_statement = key_binding_utils.handle_prettify_binding(cast(Any, mycli), statement)
    assert pretty_statement == 'SELECT\n    1;'


def test_unprettify_statement():
    statement = 'SELECT\n    1'
    mycli = FakeMyCli()
    unpretty_statement = key_binding_utils.handle_unprettify_binding(cast(Any, mycli), statement)
    assert unpretty_statement == 'SELECT 1;'


def test_handle_editor_command_returns_text_unchanged_when_not_editor_command(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(key_binding_utils.special, 'editor_command', lambda text: False)

    mycli = FakeMyCli()

    assert key_binding_utils.handle_editor_command(cast(Any, mycli), 'select 1', None, lambda: 'loaded') == 'select 1'


def test_handle_editor_command_opens_editor_reprompts_after_keyboard_interrupt_and_returns_text(monkeypatch: pytest.MonkeyPatch) -> None:
    prompt_session = FakePromptSession([KeyboardInterrupt(), 'edited sql'])
    mycli = FakeMyCli(prompt_session=prompt_session)
    open_calls: list[dict[str, str]] = []

    def inputhook(*args: object, **kwargs: object) -> None:
        return None

    def loaded_message_fn() -> str:
        return 'loaded'

    def open_external_editor(*, filename: str | None, sql: str) -> tuple[str, str | None]:
        open_calls.append({'filename': cast(str, filename), 'sql': sql})
        return 'SELECT 1', None

    monkeypatch.setattr(key_binding_utils, 'PromptSession', FakePromptSession)
    monkeypatch.setattr(key_binding_utils.special, 'editor_command', lambda text: text in {'\\e', ''})
    monkeypatch.setattr(key_binding_utils.special, 'get_filename', lambda text: 'query.sql')
    monkeypatch.setattr(key_binding_utils.special, 'get_editor_query', lambda text: '' if text == '\\e' else None)
    monkeypatch.setattr(
        key_binding_utils.special,
        'open_external_editor',
        open_external_editor,
    )

    result = key_binding_utils.handle_editor_command(cast(Any, mycli), '\\e', inputhook, loaded_message_fn)

    assert result == 'edited sql'
    assert open_calls == [{'filename': 'query.sql', 'sql': 'last query'}]
    assert prompt_session.prompt_calls == [
        {'default': 'SELECT 1', 'inputhook': inputhook, 'message': loaded_message_fn},
        {'default': '', 'inputhook': inputhook, 'message': loaded_message_fn},
    ]


def test_handle_editor_command_uses_explicit_editor_query_and_raises_on_editor_error(monkeypatch: pytest.MonkeyPatch) -> None:
    mycli = FakeMyCli(prompt_session=FakePromptSession([]))

    monkeypatch.setattr(key_binding_utils.special, 'editor_command', lambda text: True)
    monkeypatch.setattr(key_binding_utils.special, 'get_filename', lambda text: 'query.sql')
    monkeypatch.setattr(key_binding_utils.special, 'get_editor_query', lambda text: 'select from text')
    monkeypatch.setattr(
        key_binding_utils.special,
        'open_external_editor',
        lambda *, filename, sql: ('', 'editor failed'),
    )

    with pytest.raises(RuntimeError, match='editor failed'):
        key_binding_utils.handle_editor_command(cast(Any, mycli), '\\eselect 1', None, lambda: 'loaded')


def test_handle_clip_command_returns_false_when_not_clip_command(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(key_binding_utils.special, 'clip_command', lambda text: False)

    mycli = FakeMyCli()

    assert key_binding_utils.handle_clip_command(cast(Any, mycli), 'select 1') is False


def test_handle_clip_command_copies_explicit_query(monkeypatch: pytest.MonkeyPatch) -> None:
    clipboard_calls: list[str] = []

    def copy_query_to_clipboard(*, sql: str) -> None:
        clipboard_calls.append(sql)

    monkeypatch.setattr(key_binding_utils.special, 'clip_command', lambda text: True)
    monkeypatch.setattr(key_binding_utils.special, 'get_clip_query', lambda text: 'select 1')
    monkeypatch.setattr(
        key_binding_utils.special,
        'copy_query_to_clipboard',
        copy_query_to_clipboard,
    )

    mycli = FakeMyCli()

    assert key_binding_utils.handle_clip_command(cast(Any, mycli), '\\clip select 1') is True
    assert clipboard_calls == ['select 1']


def test_handle_clip_command_uses_last_query_and_raises_on_clipboard_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(key_binding_utils.special, 'clip_command', lambda text: True)
    monkeypatch.setattr(key_binding_utils.special, 'get_clip_query', lambda text: '')
    monkeypatch.setattr(
        key_binding_utils.special,
        'copy_query_to_clipboard',
        lambda *, sql: 'clipboard failed',
    )

    mycli = FakeMyCli()

    with pytest.raises(RuntimeError, match='clipboard failed'):
        key_binding_utils.handle_clip_command(cast(Any, mycli), '\\clip')


def test_prettify_statement_returns_empty_string_for_empty_input() -> None:
    mycli = FakeMyCli()
    assert key_binding_utils.handle_prettify_binding(cast(Any, mycli), '') == ''


def test_unprettify_statement_returns_empty_string_for_empty_input() -> None:
    mycli = FakeMyCli()
    assert key_binding_utils.handle_unprettify_binding(cast(Any, mycli), '') == ''


@pytest.mark.parametrize(
    ('handler_name', 'text'),
    [
        ('handle_prettify_binding', 'SELECT 1;'),
        ('handle_unprettify_binding', 'SELECT 1;'),
    ],
)
def test_prettify_helpers_fall_back_to_input_without_trailing_semicolon_on_parse_error(
    monkeypatch: pytest.MonkeyPatch,
    handler_name: str,
    text: str,
) -> None:
    monkeypatch.setattr(key_binding_utils.sqlglot, 'parse', lambda *_args, **_kwargs: (_ for _ in ()).throw(ValueError('bad sql')))

    handler = getattr(key_binding_utils, handler_name)

    mycli = FakeMyCli()

    assert handler(cast(Any, mycli), text) == 'SELECT 1'


@pytest.mark.parametrize(
    ('handler_name', 'text'),
    [
        ('handle_prettify_binding', 'SELECT 1; SELECT 2;'),
        ('handle_unprettify_binding', 'SELECT 1; SELECT 2;'),
    ],
)
def test_prettify_helpers_fall_back_when_parse_returns_multiple_statements(
    monkeypatch: pytest.MonkeyPatch,
    handler_name: str,
    text: str,
) -> None:
    monkeypatch.setattr(key_binding_utils.sqlglot, 'parse', lambda *_args, **_kwargs: [object(), object()])

    handler = getattr(key_binding_utils, handler_name)

    mycli = FakeMyCli()

    assert handler(cast(Any, mycli), text) == 'SELECT 1; SELECT 2'



"""Background prefetcher for multi-schema auto-completion.

The default completion refresher only populates metadata for the
currently-selected schema.  ``SchemaPrefetcher`` loads metadata for
additional schemas on a background thread so that users can get
qualified auto-completion suggestions (``OtherSchema.table``) without
switching databases first.
"""

from __future__ import annotations

from enum import Enum
import logging
import threading
from typing import TYPE_CHECKING, Any, Iterable

from mycli.sqlexecute import SQLExecute

if TYPE_CHECKING:  # pragma: no cover - typing only
    from mycli.client import MyCli
    from mycli.sqlcompleter import SQLCompleter

_logger = logging.getLogger(__name__)


class PrefetchMode(str, Enum):
    ALWAYS = 'always'
    NEVER = 'never'
    LISTED = 'listed'


def parse_prefetch_config(mode: str, schema_list: list[str]) -> list[str] | None:
    """Parse the ``prefetch_schemas_mode`` / ``prefetch_schemas_list`` options.

    Returns ``None`` when every accessible schema should be prefetched
    (``always``), an empty list when prefetching is disabled
    (``never``), or ``schema_list`` when the mode is ``listed``.
    Unknown modes fall back to ``always``.
    """
    try:
        parsed = PrefetchMode(mode.strip().lower())
    except ValueError:
        return None
    if parsed is PrefetchMode.NEVER:
        return []
    if parsed is PrefetchMode.LISTED:
        return schema_list
    return None


class SchemaPrefetcher:
    """Run schema prefetch work on a dedicated background thread."""

    def __init__(self, mycli: 'MyCli') -> None:
        self.mycli = mycli
        self._thread: threading.Thread | None = None
        self._cancel = threading.Event()
        self._loaded: set[str] = set()

    def is_prefetching(self) -> bool:
        return bool(self._thread and self._thread.is_alive())

    def clear_loaded(self) -> None:
        """Forget which schemas have been prefetched (used on reset)."""
        self._loaded.clear()

    def stop(self, timeout: float = 2.0) -> None:
        """Signal the background thread to stop and wait briefly for it."""
        if self._thread and self._thread.is_alive():
            self._cancel.set()
            self._thread.join(timeout=timeout)
        self._cancel = threading.Event()
        self._thread = None

    def start_configured(self) -> None:
        """Start prefetching based on the user's prefetch settings."""
        mode = getattr(self.mycli, 'prefetch_schemas_mode', PrefetchMode.ALWAYS.value)
        schema_list = getattr(self.mycli, 'prefetch_schemas_list', [])
        parsed = parse_prefetch_config(mode, schema_list)
        if parsed is not None and not parsed:
            # ``never`` or ``listed`` with an empty list — nothing to do.
            return
        self._start(parsed)

    def prefetch_schema_now(self, schema: str) -> None:
        """Fetch *schema* immediately on a background thread.

        Used when a user manually switches to a schema.  The method
        returns quickly; the actual work happens in the new thread.
        """
        if not schema:
            return
        # Avoid double-fetching while a full-prefetch pass is running.
        self.stop()
        self._start([schema])

    def _start(self, schemas: Iterable[str] | None) -> None:
        """Spawn the background worker.

        ``schemas=None`` defers resolution to the worker, which lists
        every database via its own dedicated connection — the main
        thread's ``sqlexecute`` must not be used here since the worker
        would race with the REPL.
        """
        self.stop()
        queue: list[str] | None = None if schemas is None else list(schemas)
        self._cancel = threading.Event()
        self._thread = threading.Thread(
            target=self._run,
            args=(queue,),
            name='schema_prefetcher',
            daemon=True,
        )
        self._thread.start()
        self._invalidate_app()

    def _run(self, schemas: list[str] | None) -> None:
        executor: SQLExecute | None = None
        try:
            executor = self._make_executor()
        except Exception as e:  # pragma: no cover - defensive
            _logger.error('schema prefetch could not open connection: %r', e)
            self._invalidate_app()
            return
        try:
            if schemas is None:
                try:
                    schemas = list(executor.databases())
                except Exception as e:
                    _logger.error('failed to list databases for prefetch: %r', e)
                    return
            current = self._current_schema()
            existing = set(self.mycli.completer.dbmetadata.get('tables', {}).keys())
            queue = [s for s in schemas if s and s != current and s not in self._loaded and s not in existing]
            for schema in queue:
                if self._cancel.is_set():
                    return
                try:
                    self._prefetch_one(executor, schema)
                    self._loaded.add(schema)
                except Exception as e:
                    _logger.error('prefetch failed for schema %r: %r', schema, e)
        finally:
            try:
                executor.close()
            except Exception:  # pragma: no cover - defensive
                pass
            self._invalidate_app()

    def _prefetch_one(self, executor: SQLExecute, schema: str) -> None:
        _logger.debug('prefetching schema %r', schema)
        table_rows = list(executor.table_columns(schema=schema))
        fk_rows = list(executor.foreign_keys(schema=schema))
        enum_rows = list(executor.enum_values(schema=schema))
        func_rows = list(executor.functions(schema=schema))
        proc_rows = list(executor.procedures(schema=schema))

        # Use the live completer's escape logic so keys match what the
        # completion engine computes when parsing user input.
        completer = self.mycli.completer
        table_columns: dict[str, list[str]] = {}
        for table, column in table_rows:
            esc_table = completer.escape_name(table)
            esc_col = completer.escape_name(column)
            cols = table_columns.setdefault(esc_table, ['*'])
            cols.append(esc_col)

        fk_tables: dict[str, set[str]] = {}
        fk_relations: list[tuple[str, str, str, str]] = []
        for table, col, ref_table, ref_col in fk_rows:
            esc_table = completer.escape_name(table)
            esc_col = completer.escape_name(col)
            esc_ref_table = completer.escape_name(ref_table)
            esc_ref_col = completer.escape_name(ref_col)
            fk_tables.setdefault(esc_table, set()).add(esc_ref_table)
            fk_tables.setdefault(esc_ref_table, set()).add(esc_table)
            fk_relations.append((esc_table, esc_col, esc_ref_table, esc_ref_col))
        fk_payload: dict[str, Any] = {'tables': fk_tables, 'relations': fk_relations}

        enum_values: dict[str, dict[str, list[str]]] = {}
        for table, column, values in enum_rows:
            esc_table = completer.escape_name(table)
            esc_col = completer.escape_name(column)
            enum_values.setdefault(esc_table, {})[esc_col] = list(values)

        functions: dict[str, None] = {}
        for row in func_rows:
            if not row or not row[0]:
                continue
            functions[completer.escape_name(row[0])] = None

        procedures: dict[str, None] = {}
        for row in proc_rows:
            if not row or not row[0]:
                continue
            procedures[completer.escape_name(row[0])] = None

        with self.mycli._completer_lock:
            live_completer: 'SQLCompleter' = self.mycli.completer
            live_completer.load_schema_metadata(
                schema=schema,
                table_columns=table_columns,
                foreign_keys=fk_payload,
                enum_values=enum_values,
                functions=functions,
                procedures=procedures,
            )
        self._invalidate_app()

    def _current_schema(self) -> str | None:
        sqlexecute = self.mycli.sqlexecute
        return sqlexecute.dbname if sqlexecute is not None else None

    def _make_executor(self) -> SQLExecute:
        sqlexecute = self.mycli.sqlexecute
        assert sqlexecute is not None
        return SQLExecute(
            sqlexecute.dbname,
            sqlexecute.user,
            sqlexecute.password,
            sqlexecute.host,
            sqlexecute.port,
            sqlexecute.socket,
            sqlexecute.character_set,
            sqlexecute.local_infile,
            sqlexecute.ssl,
            sqlexecute.ssh_user,
            sqlexecute.ssh_host,
            sqlexecute.ssh_port,
            sqlexecute.ssh_password,
            sqlexecute.ssh_key_filename,
        )

    def _invalidate_app(self) -> None:
        prompt_session = getattr(self.mycli, 'prompt_session', None)
        if prompt_session is None:
            return
        try:
            prompt_session.app.invalidate()
        except Exception:  # pragma: no cover - defensive
            pass



# type: ignore

from types import SimpleNamespace
from unittest.mock import MagicMock

from prompt_toolkit.enums import EditingMode
from prompt_toolkit.key_binding.vi_state import InputMode
import pytest

from mycli import clitoolbar


def make_mycli(
    *,
    smart_completion: bool = True,
    multi_line: bool = False,
    editing_mode: EditingMode = EditingMode.EMACS,
    toolbar_error_message: str | None = None,
    refreshing: bool = False,
    prefetching: bool = False,
):
    return SimpleNamespace(
        completer=SimpleNamespace(smart_completion=smart_completion),
        multi_line=multi_line,
        prompt_session=SimpleNamespace(editing_mode=editing_mode),
        toolbar_error_message=toolbar_error_message,
        completion_refresher=SimpleNamespace(is_refreshing=MagicMock(return_value=refreshing)),
        schema_prefetcher=SimpleNamespace(is_prefetching=MagicMock(return_value=prefetching)),
        get_custom_toolbar=MagicMock(return_value="custom toolbar"),
    )


def test_create_toolbar_tokens_func_shows_initial_help() -> None:
    mycli = make_mycli()

    toolbar = clitoolbar.create_toolbar_tokens_func(mycli, lambda: True, None, mycli.get_custom_toolbar)
    result = toolbar()

    assert ("class:bottom-toolbar", "right-arrow accepts full-line suggestion") in result
    assert ("class:bottom-toolbar", "[F2] Smart-complete:") in result
    assert ("class:bottom-toolbar.on", "ON ") in result
    assert ("class:bottom-toolbar", "[F3] Multiline:") in result
    assert ("class:bottom-toolbar.off", "OFF") in result


def test_create_toolbar_tokens_func_clears_toolbar_error_message() -> None:
    mycli = make_mycli(toolbar_error_message="boom")

    toolbar = clitoolbar.create_toolbar_tokens_func(mycli, lambda: False, None, mycli.get_custom_toolbar)
    first = toolbar()
    second = toolbar()

    assert ("class:bottom-toolbar.transaction.failed", "boom") in first
    assert ("class:bottom-toolbar.transaction.failed", "boom") not in second
    assert mycli.toolbar_error_message is None
    assert ("class:bottom-toolbar", "right-arrow accepts full-line suggestion") not in first


def test_create_toolbar_tokens_func_shows_prefetching() -> None:
    mycli = make_mycli(prefetching=True)

    toolbar = clitoolbar.create_toolbar_tokens_func(mycli, lambda: False, None, mycli.get_custom_toolbar)
    result = toolbar()

    assert ("class:bottom-toolbar", "Prefetching schemas…") in result


def test_create_toolbar_tokens_func_shows_multiline_vi_and_refreshing(monkeypatch) -> None:
    mycli = make_mycli(
        smart_completion=False,
        multi_line=True,
        editing_mode=EditingMode.VI,
        refreshing=True,
    )
    monkeypatch.setattr(clitoolbar.special, 'get_current_delimiter', lambda: '$$')
    monkeypatch.setattr(clitoolbar, '_get_vi_mode', lambda: 'N')

    toolbar = clitoolbar.create_toolbar_tokens_func(mycli, lambda: False, None, mycli.get_custom_toolbar)
    result = toolbar()

    assert ("class:bottom-toolbar.off", "OFF") in result
    assert ("class:bottom-toolbar", "[F3] Multiline:") in result
    assert ("class:bottom-toolbar.on", "ON ") in result
    assert ("class:bottom-toolbar", "Vi:") in result
    assert ("class:bottom-toolbar.on", "N") in result
    assert ('class:bottom-toolbar.on', '$$') in result
    assert ("class:bottom-toolbar", "Refreshing completions…") in result


def test_create_toolbar_tokens_func_applies_custom_format(monkeypatch) -> None:
    mycli = make_mycli(multi_line=True, refreshing=True)
    monkeypatch.setattr(clitoolbar.special, 'get_current_delimiter', lambda: '$$')

    formatted = [("class:bottom-toolbar", "CUSTOM")]
    to_formatted_text = MagicMock(return_value=formatted)
    monkeypatch.setattr(clitoolbar, 'to_formatted_text', to_formatted_text)

    toolbar = clitoolbar.create_toolbar_tokens_func(mycli, lambda: True, r'\Bfmt', mycli.get_custom_toolbar)
    result = toolbar()

    mycli.get_custom_toolbar.assert_called_once_with('fmt')
    to_formatted_text.assert_called_once_with("custom toolbar", style='class:bottom-toolbar')
    assert ('class:bottom-toolbar', '\n') in result
    assert ("class:bottom-toolbar", "CUSTOM") in result
    assert ("class:bottom-toolbar", "right-arrow accepts full-line suggestion") in result
    assert ("class:bottom-toolbar", "Refreshing completions…") in result


def test_create_toolbar_tokens_func_replaces_default_toolbar_for_plain_custom_format(monkeypatch) -> None:
    mycli = make_mycli(multi_line=True, toolbar_error_message='boom', refreshing=True)
    monkeypatch.setattr(clitoolbar.special, 'get_current_delimiter', lambda: '$$')

    formatted = [('class:bottom-toolbar', 'PLAIN CUSTOM')]
    to_formatted_text = MagicMock(return_value=formatted)
    monkeypatch.setattr(clitoolbar, 'to_formatted_text', to_formatted_text)

    toolbar = clitoolbar.create_toolbar_tokens_func(mycli, lambda: True, 'fmt', mycli.get_custom_toolbar)
    result = toolbar()

    mycli.get_custom_toolbar.assert_called_once_with('fmt')
    to_formatted_text.assert_called_once_with('custom toolbar', style='class:bottom-toolbar')
    assert ('class:bottom-toolbar', 'PLAIN CUSTOM') in result
    assert ('class:bottom-toolbar', '[Tab] Complete') not in result
    assert ('class:bottom-toolbar', '[F1] Help') not in result
    assert ('class:bottom-toolbar', 'right-arrow accepts full-line suggestion') in result
    assert ('class:bottom-toolbar.transaction.failed', 'boom') in result


@pytest.mark.parametrize(
    ('input_mode', 'expected'),
    [
        (InputMode.INSERT, 'I'),
        (InputMode.NAVIGATION, 'N'),
        (InputMode.REPLACE, 'R'),
        (InputMode.REPLACE_SINGLE, 'R'),
        (InputMode.INSERT_MULTIPLE, 'M'),
    ],
)
def test_get_vi_mode(monkeypatch, input_mode: InputMode, expected: str) -> None:
    app = SimpleNamespace(vi_state=SimpleNamespace(input_mode=input_mode))
    monkeypatch.setattr(clitoolbar, 'get_app', lambda: app)

    assert clitoolbar._get_vi_mode() == expected



# type: ignore

from textwrap import dedent

from behave import then, when
from utils import parse_cli_args_to_dict
import wrappers


@when("we run dbcli with {arg}")
def step_run_cli_with_arg(context, arg):
    wrappers.run_cli(context, run_args=parse_cli_args_to_dict(arg))


@when("we execute a small query")
def step_execute_small_query(context):
    context.cli.sendline("select 1")


@when("we execute a large query")
def step_execute_large_query(context):
    context.cli.sendline(f"select {','.join([str(n) for n in range(1, 50)])}")


@then("we see small results in horizontal format")
def step_see_small_results(context):
    expected = (
        dedent(
            """
            +---+\r
            | 1 |\r
            +---+\r
            | 1 |\r
            +---+
            """
        ).strip()
        + '\r\n\r\n'
    )

    wrappers.expect_pager(
        context,
        expected,
        timeout=5,
    )
    wrappers.expect_exact(context, "1 row in set", timeout=2)


@then("we see large results in vertical format")
def step_see_large_results(context):
    rows = [f"{str(n):3}| {n}" for n in range(1, 50)]
    delimited_rows = '\r\n'.join(rows) + '\r\n'
    expected = "***************************[ 1. row ]***************************\r\n" + delimited_rows + "\r\n"

    wrappers.expect_pager(context, expected, timeout=10)
    wrappers.expect_exact(context, "1 row in set", timeout=2)



import datetime
import logging
import os
from typing import Any

import click
import pymysql
from pymysql.cursors import Cursor

logger = logging.getLogger(__name__)

CACHED_SSL_VERSION: dict[tuple, str | None] = {}


def handle_cd_command(command: list[str]) -> tuple[bool, str | None]:
    """Handles a `cd` shell command by calling python's os.chdir."""
    if not command[0].lower() == 'cd':
        return False, 'Not a cd command.'
    if len(command) != 2:
        return False, 'Exactly one directory name must be provided.'
    directory = command[1]
    try:
        os.chdir(directory)
        click.echo(os.getcwd(), err=True)
        return True, None
    except OSError as e:
        return False, e.strerror


def format_uptime(uptime_in_seconds: str) -> str:
    """Format number of seconds into human-readable string.

    :param uptime_in_seconds: The server uptime in seconds.
    :returns: A human-readable string representing the uptime.

    >>> uptime = format_uptime('56892')
    >>> print(uptime)
    15 hours 48 min 12 sec
    """

    m, s = divmod(int(uptime_in_seconds), 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)

    uptime_values: list[str] = []

    for value, unit in ((d, "days"), (h, "hours"), (m, "min"), (s, "sec")):
        if value == 0 and not uptime_values:
            # Don't include a value/unit if the unit isn't applicable to
            # the uptime. E.g. don't do 0 days 0 hours 1 min 30 sec.
            continue
        if value == 1 and unit.endswith("s"):
            # Remove the "s" if the unit is singular.
            unit = unit[:-1]
        uptime_values.append(f'{value} {unit}')

    uptime = " ".join(uptime_values)
    return uptime


def get_uptime(cur: Cursor) -> int:
    query = 'SHOW STATUS LIKE "Uptime"'
    logger.debug(query)

    uptime = 0

    try:
        cur.execute(query)
        if one := cur.fetchone():
            uptime = int(one[1] or 0)
    except pymysql.err.OperationalError:
        pass

    return uptime


def get_warning_count(cur: Cursor) -> int:
    query = 'SHOW COUNT(*) WARNINGS'
    logger.debug(query)

    warning_count = 0

    try:
        cur.execute(query)
        if one := cur.fetchone():
            warning_count = int(one[0] or 0)
    except pymysql.err.OperationalError:
        pass

    return warning_count


def get_ssl_version(cur: Cursor) -> str | None:
    cache_key = (id(cur.connection), cur.connection.thread_id())

    if cache_key in CACHED_SSL_VERSION:
        return CACHED_SSL_VERSION[cache_key] or None

    query = 'SHOW STATUS LIKE "Ssl_version"'
    logger.debug(query)

    ssl_version = None

    try:
        cur.execute(query)
        if one := cur.fetchone():
            CACHED_SSL_VERSION[cache_key] = one[1]
            ssl_version = one[1] or None
        else:
            CACHED_SSL_VERSION[cache_key] = ''
    except pymysql.err.OperationalError:
        pass

    return ssl_version


def get_ssl_cipher(cur: Cursor) -> str | None:
    query = 'SHOW STATUS LIKE "Ssl_cipher"'
    logger.debug(query)

    ssl_cipher = None

    try:
        cur.execute(query)
        if one := cur.fetchone():
            ssl_cipher = one[1] or None
    except pymysql.err.OperationalError:
        pass

    return ssl_cipher


def get_server_timezone(variables: dict[str, Any]) -> str:
    try:
        if variables['time_zone'] == 'SYSTEM':
            server_tz = variables['system_time_zone']
        else:
            server_tz = variables['time_zone']
        return server_tz
    except KeyError:
        return ''


def get_local_timezone() -> str:
    return datetime.datetime.now().astimezone().tzname() or ''



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from mycli.client import MyCli
    from mycli.main import CliArgs


def main_execute_from_cli(mycli: 'MyCli', cli_args: 'CliArgs') -> int:
    if cli_args.execute is None:
        return 1
    if not sys.stdin.isatty():
        click.secho('Ignoring STDIN since --execute was also given.', err=True, fg='red')
    if cli_args.batch:
        click.secho('Ignoring --batch since --execute was also given.', err=True, fg='red')
    try:
        execute_sql = cli_args.execute
        if cli_args.format == 'csv':
            mycli.main_formatter.format_name = 'csv'
            if execute_sql.endswith(r'\G'):
                execute_sql = execute_sql[:-2]
        elif cli_args.format == 'tsv':
            mycli.main_formatter.format_name = 'tsv'
            if execute_sql.endswith(r'\G'):
                execute_sql = execute_sql[:-2]
        elif cli_args.format == 'table':
            mycli.main_formatter.format_name = 'ascii'
            if execute_sql.endswith(r'\G'):
                execute_sql = execute_sql[:-2]
        else:
            mycli.main_formatter.format_name = 'tsv'

        mycli.run_query(execute_sql, checkpoint=cli_args.checkpoint)
        return 0
    except Exception as e:
        click.secho(str(e), err=True, fg="red")
        return 1



# type: ignore

import pytest

from mycli.packages import cli_utils
from mycli.packages.cli_utils import (
    filtered_sys_argv,
    is_valid_connection_scheme,
)


@pytest.mark.parametrize(
    ('argv', 'expected'),
    [
        (['mycli', '-h'], ['--help']),
        (['mycli', '-h', 'example.com'], ['-h', 'example.com']),
    ],
)
def test_filtered_sys_argv(monkeypatch, argv, expected):
    monkeypatch.setattr(cli_utils.sys, 'argv', argv)

    assert filtered_sys_argv() == expected


@pytest.mark.parametrize(
    ('text', 'is_valid', 'invalid_scheme'),
    [
        ('localhost', False, None),
        ('mysql://user@localhost/db', True, None),
        ('mysqlx://user@localhost/db', True, None),
        ('tcp://localhost:3306', True, None),
        ('socket:///tmp/mysql.sock', True, None),
        ('ssh://user@example.com', True, None),
        ('postgres://user@localhost/db', False, 'postgres'),
        ('http://example.com', False, 'http'),
    ],
)
def test_is_valid_connection_scheme(text, is_valid, invalid_scheme):
    assert is_valid_connection_scheme(text) == (is_valid, invalid_scheme)



# type: ignore

import os
import shutil
import sys
from tempfile import NamedTemporaryFile

import db_utils as dbutils
import fixture_utils as fixutils
import pexpect

from mycli.constants import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_USER
from steps.wrappers import run_cli, wait_prompt
from test.utils import TEMPFILE_PREFIX

test_log_file = os.path.join(os.environ["HOME"], ".mycli.test.log")


SELF_CONNECTING_FEATURES = ("test/features/connection.feature",)


MY_CNF_PATH = os.path.expanduser("~/.my.cnf")
MY_CNF_BACKUP_PATH = f"{MY_CNF_PATH}.backup"
MYLOGIN_CNF_PATH = os.path.expanduser("~/.mylogin.cnf")
MYLOGIN_CNF_BACKUP_PATH = f"{MYLOGIN_CNF_PATH}.backup"


def get_db_name_from_context(context):
    return context.config.userdata.get("my_test_db", None) or "mycli_behave_tests"


def before_all(context):
    """Set env parameters."""
    os.environ["LINES"] = "100"
    os.environ["COLUMNS"] = "100"
    os.environ["VISUAL"] = "ex"
    os.environ["EDITOR"] = "ex"
    os.environ["LC_ALL"] = "en_US.UTF-8"
    os.environ["PROMPT_TOOLKIT_NO_CPR"] = "1"
    os.environ["MYCLI_HISTFILE"] = os.devnull

    # test_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
    # login_path_file = os.path.join(test_dir, "mylogin.cnf")
    #    os.environ['MYSQL_TEST_LOGIN_FILE'] = login_path_file

    context.package_root = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
    context.exit_sent = False

    vi = "_".join([str(x) for x in sys.version_info[:3]])
    db_name = get_db_name_from_context(context)
    db_name_full = f"{db_name}_{vi}"

    # Store get params from config/environment variables
    context.conf = {
        "host": context.config.userdata.get("my_test_host", os.getenv("PYTEST_HOST", DEFAULT_HOST)),
        "port": context.config.userdata.get("my_test_port", int(os.getenv("PYTEST_PORT", DEFAULT_PORT))),
        "user": context.config.userdata.get("my_test_user", os.getenv("PYTEST_USER", DEFAULT_USER)),
        "pass": context.config.userdata.get("my_test_pass", os.getenv("PYTEST_PASSWORD", None)),
        "cli_command": context.config.userdata.get("my_cli_command", None)
        or sys.executable + ' -c "import coverage ; coverage.process_startup(); import mycli.main; mycli.main.click_entrypoint()"',
        "dbname": db_name,
        "dbname_tmp": db_name_full + "_tmp",
        "vi": vi,
        "pager_boundary": "---boundary---",
    }

    with NamedTemporaryFile(prefix=TEMPFILE_PREFIX, mode='w', delete=False) as my_cnf:
        my_cnf.write(
            f'[client]\npager={sys.executable} '
            f'{os.path.join(context.package_root, "test/features/wrappager.py")} {context.conf["pager_boundary"]}\n'
        )
    context.conf["defaults-file"] = my_cnf.name
    context.conf["myclirc"] = os.path.join(context.package_root, "test", "myclirc")

    context.cn = dbutils.create_db(
        context.conf["host"], context.conf["port"], context.conf["user"], context.conf["pass"], context.conf["dbname"]
    )

    context.fixture_data = fixutils.read_fixture_files()


def after_all(context):
    """Unset env parameters."""
    dbutils.close_cn(context.cn)
    dbutils.drop_db(context.conf["host"], context.conf["port"], context.conf["user"], context.conf["pass"], context.conf["dbname"])
    try:
        if os.path.exists(context.conf["defaults-file"]):
            os.remove(context.conf["defaults-file"])
    except Exception:
        pass

    # Restore env vars.
    # for k, v in context.pgenv.items():
    #    if k in os.environ and v is None:
    #        del os.environ[k]
    #    elif v:
    #        os.environ[k] = v


def before_step(context, _):
    context.atprompt = False


def before_scenario(context, arg):
    # Skip scenarios marked skip_py312 when running on Python 3.12
    if sys.version_info[:2] == (3, 12) and "skip_py312" in arg.tags:
        arg.skip("Skipped on Python 3.12")
    # Skip flaky editor test in CI
    if os.getenv('GITHUB_ACTION') and 'skip_ci' in arg.tags:
        arg.skip('Skipped in CI')
    with open(test_log_file, "w") as f:
        f.write("")
    if arg.location.filename not in SELF_CONNECTING_FEATURES:
        run_cli(context)
        wait_prompt(context)

    if os.path.exists(MY_CNF_PATH):
        shutil.move(MY_CNF_PATH, MY_CNF_BACKUP_PATH)

    if os.path.exists(MYLOGIN_CNF_PATH):
        shutil.move(MYLOGIN_CNF_PATH, MYLOGIN_CNF_BACKUP_PATH)


def after_scenario(context, _):
    """Cleans up after each test complete."""
    with open(test_log_file) as f:
        for line in f:
            if "error" in line.lower():
                raise RuntimeError(f"Error in log file: {line}")

    if hasattr(context, "cli") and not context.exit_sent:
        # Quit nicely.
        if not context.atprompt:
            user = context.conf["user"]
            host = context.conf["host"]
            dbname = context.currentdb
            context.cli.expect_exact(f"{user}@{host}:{dbname}>", timeout=5)
        context.cli.sendcontrol("c")
        context.cli.sendcontrol("d")
        context.cli.expect_exact(pexpect.EOF, timeout=5)

    if os.path.exists(MY_CNF_BACKUP_PATH):
        shutil.move(MY_CNF_BACKUP_PATH, MY_CNF_PATH)

    if os.path.exists(MYLOGIN_CNF_BACKUP_PATH):
        shutil.move(MYLOGIN_CNF_BACKUP_PATH, MYLOGIN_CNF_PATH)
    elif os.path.exists(MYLOGIN_CNF_PATH):
        # This file was moved in `before_scenario`.
        # If it exists now, it has been created during a test
        os.remove(MYLOGIN_CNF_PATH)


# TODO: uncomment to debug a failure
# def after_step(context, step):
#     if step.status == "failed":
#         import ipdb; ipdb.set_trace()



# type: ignore

"""Test the sql output adapter."""

import os
from textwrap import dedent

from cli_helpers.utils import strip_ansi
from pymysql.constants import FIELD_TYPE
import pytest

from mycli.main import MyCli
from mycli.packages.sqlresult import SQLResult
from test.utils import HOST, PASSWORD, PORT, USER, dbtest

default_config_file = os.path.join(os.path.dirname(__file__), "../myclirc")


@pytest.fixture
def mycli():
    cli = MyCli()
    cli.connect(None, USER, PASSWORD, HOST, PORT, None, init_command=None)
    yield cli
    cli.sqlexecute.conn.close()


@dbtest
def test_sql_output(mycli):
    """Test the sql output adapter."""
    header = ["letters", "number", "optional", "float", "binary"]

    class FakeCursor:
        def __init__(self):
            self.data = [("abc", 1, None, 10.0, b"\xaa"), ("d", 456, "1", 0.5, b"\xaa\xbb")]
            self.description = [
                (None, FIELD_TYPE.VARCHAR),
                (None, FIELD_TYPE.LONG),
                (None, FIELD_TYPE.LONG),
                (None, FIELD_TYPE.FLOAT),
                (None, FIELD_TYPE.BLOB),
            ]

        def __iter__(self):
            return self

        def __next__(self):
            if self.data:
                return self.data.pop(0)
            else:
                raise StopIteration()

        def description(self):
            return self.description

    # Test sql-update output format
    assert list(mycli.change_table_format("sql-update")) == [SQLResult(status="Changed table format to sql-update")]
    mycli.main_formatter.query = ""
    mycli.redirect_formatter.query = ""
    output = mycli.format_sqlresult(SQLResult(header=header, rows=FakeCursor()))
    actual = "\n".join(output)
    assert actual == dedent("""\
            UPDATE `DUAL` SET
              `number` = 1
            , `optional` = NULL
            , `float` = 10.0e0
            , `binary` = 0xaa
            WHERE `letters` = 'abc';
            UPDATE `DUAL` SET
              `number` = 456
            , `optional` = '1'
            , `float` = 0.5e0
            , `binary` = 0xaabb
            WHERE `letters` = 'd';""")
    # Test sql-update-2 output format
    assert list(mycli.change_table_format("sql-update-2")) == [SQLResult(status="Changed table format to sql-update-2")]
    mycli.main_formatter.query = ""
    mycli.redirect_formatter.query = ""
    output = mycli.format_sqlresult(SQLResult(header=header, rows=FakeCursor()))
    assert "\n".join(output) == dedent("""\
            UPDATE `DUAL` SET
              `optional` = NULL
            , `float` = 10.0e0
            , `binary` = 0xaa
            WHERE `letters` = 'abc' AND `number` = 1;
            UPDATE `DUAL` SET
              `optional` = '1'
            , `float` = 0.5e0
            , `binary` = 0xaabb
            WHERE `letters` = 'd' AND `number` = 456;""")
    # Test sql-insert output format (without table name)
    assert list(mycli.change_table_format("sql-insert")) == [SQLResult(status="Changed table format to sql-insert")]
    mycli.main_formatter.query = ""
    mycli.redirect_formatter.query = ""
    output = mycli.format_sqlresult(SQLResult(header=header, rows=FakeCursor()))
    assert "\n".join(output) == dedent("""\
            INSERT INTO `DUAL` (`letters`, `number`, `optional`, `float`, `binary`) VALUES
              ('abc', 1, NULL, 10.0e0, 0xaa)
            , ('d', 456, '1', 0.5e0, 0xaabb)
            ;""")
    # Test sql-insert output format (with table name)
    assert list(mycli.change_table_format("sql-insert")) == [SQLResult(status="Changed table format to sql-insert")]
    mycli.main_formatter.query = "SELECT * FROM `table`"
    mycli.redirect_formatter.query = "SELECT * FROM `table`"
    output = mycli.format_sqlresult(SQLResult(header=header, rows=FakeCursor()))
    assert "\n".join(output) == dedent("""\
            INSERT INTO table (`letters`, `number`, `optional`, `float`, `binary`) VALUES
              ('abc', 1, NULL, 10.0e0, 0xaa)
            , ('d', 456, '1', 0.5e0, 0xaabb)
            ;""")
    # Test sql-insert output format (with database + table name)
    assert list(mycli.change_table_format("sql-insert")) == [SQLResult(status="Changed table format to sql-insert")]
    mycli.main_formatter.query = "SELECT * FROM `database`.`table`"
    mycli.redirect_formatter.query = "SELECT * FROM `database`.`table`"
    output = mycli.format_sqlresult(SQLResult(header=header, rows=FakeCursor()))
    assert "\n".join(output) == dedent("""\
            INSERT INTO database.table (`letters`, `number`, `optional`, `float`, `binary`) VALUES
              ('abc', 1, NULL, 10.0e0, 0xaa)
            , ('d', 456, '1', 0.5e0, 0xaabb)
            ;""")
    # Test binary output format is a hex string
    assert list(mycli.change_table_format("psql")) == [SQLResult(status="Changed table format to psql")]
    output = mycli.format_sqlresult(SQLResult(header=header, rows=FakeCursor()))
    assert '0xaabb' in '\n'.join(output)


@dbtest
def test_postamble_output(mycli):
    """Test the postamble output property."""
    header = ['letters', 'number', 'optional', 'float']

    class FakeCursor:
        def __init__(self):
            self.data = [('abc', 1, None, 10.0)]
            self.description = [
                (None, FIELD_TYPE.VARCHAR),
                (None, FIELD_TYPE.LONG),
                (None, FIELD_TYPE.LONG),
                (None, FIELD_TYPE.FLOAT),
            ]

        def __iter__(self):
            return self

        def __next__(self):
            if self.data:
                return self.data.pop(0)
            else:
                raise StopIteration()

        def description(self):
            return self.description

    postamble = 'postamble:\nfooter content'
    mycli.change_table_format('ascii')
    mycli.main_formatter.query = ''
    output = mycli.format_sqlresult(SQLResult(header=header, rows=FakeCursor(), postamble=postamble))
    actual = "\n".join(output)
    assert actual.endswith(postamble)


def test_tabulate_output_preserves_multiline_whitespace(monkeypatch, tmp_path):
    monkeypatch.setenv("HOME", str(tmp_path))
    mycli = MyCli(myclirc=default_config_file)
    mycli.helpers_style = None
    mycli.helpers_warnings_style = None

    assert list(mycli.change_table_format("ascii")) == [SQLResult(status="Changed table format to ascii")]

    output = mycli.format_sqlresult(SQLResult(header=["text"], rows=[["  one\n       two\nthree"]]))

    assert strip_ansi("\n".join(output)) == dedent("""\
        +------------+
        | text       |
        +------------+
        |   one      |
        |        two |
        | three      |
        +------------+""")



import functools
import logging

import sqlglot

from mycli.compat import WIN
from mycli.packages.special.delimitercommand import DelimiterCommand

logger = logging.getLogger(__name__)
delimiter_command = DelimiterCommand()


def find_token_indices(tokens: list[sqlglot.Token]) -> dict[str, list[int]]:
    token_indices: dict[str, list[int]] = {
        'raw_dollar': [],
        'true_dollar': [],
        'angle_bracket': [],
        'pipe': [],
    }

    for i, tok in enumerate(tokens):
        if tok.token_type == sqlglot.TokenType.VAR and tok.text == '$':
            token_indices['raw_dollar'].append(i)
            continue
        if tok.token_type == sqlglot.TokenType.GT and (i - 1) in token_indices['raw_dollar']:
            token_indices['angle_bracket'].append(i)
            continue
        if tok.token_type == sqlglot.TokenType.PIPE and (i - 1) in token_indices['raw_dollar']:
            token_indices['pipe'].append(i)
            continue

    for i in token_indices['raw_dollar']:
        if (i + 1) in token_indices['angle_bracket'] or (i + 1) in token_indices['pipe']:
            token_indices['true_dollar'].append(i)

    return token_indices


def find_sql_part(
    command: str,
    tokens: list[sqlglot.Token],
    true_dollar_indices: list[int],
):
    leftmost_dollar_pos = tokens[true_dollar_indices[0]].start
    sql_part = command[0:leftmost_dollar_pos].strip().removesuffix(delimiter_command.current).rstrip()
    try:
        statements = sqlglot.parse(sql_part, read='mysql')
    except sqlglot.errors.ParseError:
        return ''
    if len(statements) != 1:
        # buglet: the statement count doesn't respect a custom delimiter
        return ''
    return sql_part


def find_command_tokens(
    tokens: list[sqlglot.Token],
    true_dollar_indices: list[int],
) -> list[sqlglot.Token]:
    command_part_tokens = []

    for i, tok in enumerate(tokens):
        if i < true_dollar_indices[0]:
            continue
        if i in true_dollar_indices:
            continue
        command_part_tokens.append(tok)

    if command_part_tokens:
        _operator = command_part_tokens.pop(0)

    return command_part_tokens


def find_file_tokens(
    tokens: list[sqlglot.Token],
    angle_bracket_indices: list[int],
) -> tuple[list[sqlglot.Token], int, str | None]:
    file_part_tokens: list[sqlglot.Token] = []
    file_part_index = len(tokens)

    if not angle_bracket_indices:
        return file_part_tokens, file_part_index, None

    file_part_tokens = tokens[angle_bracket_indices[-1] :]
    file_part_index = angle_bracket_indices[-1]

    file_operator_part = file_part_tokens.pop(0).text
    if file_operator_part == '>' and file_part_tokens[0].token_type == sqlglot.TokenType.GT:
        file_part_tokens.pop(0)
        file_operator_part = '>>'

    return file_part_tokens, file_part_index, file_operator_part


def assemble_tokens(tokens: list[sqlglot.Token]) -> str:
    assembled_string = ' ' * (tokens[-1].end + 10)
    for tok in tokens:
        if tok.token_type == sqlglot.TokenType.IDENTIFIER:
            text = f'"{tok.text}"'
            offset = 2
        elif tok.token_type == sqlglot.TokenType.STRING:
            text = f"'{tok.text}'"
            offset = 2
        else:
            text = tok.text
            offset = 0
        assembled_string = assembled_string[0 : tok.start] + text + assembled_string[tok.end + offset :]
    return assembled_string.strip().removesuffix(delimiter_command.current).rstrip()


def invalid_shell_part(
    file_part: str | None,
    command_part: str | None,
) -> bool:
    if file_part and ' ' in file_part:
        return True

    if file_part and '>' in file_part:
        return True

    if not file_part and not command_part:
        return True

    return False


# todo there are still corner cases combining custom delimiters, caching, and redirection
@functools.lru_cache(maxsize=1)
def get_redirect_components(command: str) -> tuple[str | None, str | None, str | None, str | None]:
    """Get the parts of a hybrid shell-style redirect command."""

    try:
        tokens = sqlglot.tokenize(command)
    except sqlglot.errors.TokenError:
        return None, None, None, None

    token_indices = find_token_indices(tokens)

    if not token_indices['true_dollar']:
        return None, None, None, None

    if len(token_indices['angle_bracket']) > 1:
        return None, None, None, None

    if WIN and len(token_indices['pipe']) > 1:
        # how to give better feedback here?
        return None, None, None, None

    if token_indices['angle_bracket'] and token_indices['pipe']:
        if token_indices['pipe'][-1] > token_indices['angle_bracket'][-1]:
            return None, None, None, None

    sql_part = find_sql_part(
        command,
        tokens,
        token_indices['true_dollar'],
    )
    if not sql_part:
        return None, None, None, None

    (
        file_part_tokens,
        file_part_index,
        file_operator_part,
    ) = find_file_tokens(
        tokens,
        token_indices['angle_bracket'],
    )

    command_part_tokens = find_command_tokens(
        tokens[0:file_part_index],
        token_indices['true_dollar'],
    )

    if file_part_tokens:
        file_part = assemble_tokens(file_part_tokens)
    else:
        file_part = None

    if command_part_tokens:
        command_part = assemble_tokens(command_part_tokens)
    else:
        command_part = None

    if invalid_shell_part(file_part, command_part):
        return None, None, None, None

    logger.debug('redirect parse sql_part: "{}"'.format(sql_part))
    logger.debug('redirect parse command_part: "{}"'.format(command_part))
    logger.debug('redirect parse file_operator_part: "{}"'.format(file_operator_part))
    logger.debug('redirect parse file_part: "{}"'.format(file_part))

    return sql_part, command_part, file_operator_part, file_part


def is_redirect_command(command: str) -> bool:
    """Is this a shell-style redirect to command or file?

    :param command: string

    """
    sql_part, _command_part, _file_operator_part, _file_part = get_redirect_components(command)
    return bool(sql_part)



import sys

import click

from mycli.packages.sql_utils import is_destructive


class ConfirmBoolParamType(click.ParamType):
    name = "confirmation"

    def convert(self, value: bool | str, param: click.Parameter | None, ctx: click.Context | None) -> bool:
        if isinstance(value, bool):
            return bool(value)
        value = value.lower()
        if value in ("yes", "y"):
            return True
        if value in ("no", "n"):
            return False
        self.fail(f'{value} is not a valid boolean', param, ctx)

    def __repr__(self):
        return "BOOL"


BOOLEAN_TYPE = ConfirmBoolParamType()


def confirm_destructive_query(keywords: list[str], queries: str) -> bool | None:
    """Check if the query is destructive and prompts the user to confirm.

    Returns:
    * None if the query is non-destructive or we can't prompt the user.
    * True if the query is destructive and the user wants to proceed.
    * False if the query is destructive and the user doesn't want to proceed.

    """
    prompt_text = "You're about to run a destructive command.\nDo you want to proceed? (y/n)"
    if is_destructive(keywords, queries) and sys.stdin.isatty():
        return prompt(prompt_text, type=BOOLEAN_TYPE)
    else:
        return None


def confirm(*args, **kwargs) -> bool:
    """Prompt for confirmation (yes/no) and handle any abort exceptions."""
    try:
        return click.confirm(*args, **kwargs)
    except click.Abort:
        return False


def prompt(*args, **kwargs) -> bool:
    """Prompt the user for input and handle any abort exceptions."""
    try:
        return click.prompt(*args, **kwargs)
    except click.Abort:
        return False



from __future__ import annotations

from dataclasses import dataclass
from types import SimpleNamespace
from typing import Any, cast

import pytest

import mycli.main_modes.execute as execute_mode


@dataclass
class DummyCliArgs:
    execute: str | None
    format: str = 'tsv'
    batch: str | None = None
    checkpoint: str | None = None


@dataclass
class DummyFormatter:
    format_name: str | None = None


class DummyMyCli:
    def __init__(self, run_query_error: Exception | None = None) -> None:
        self.main_formatter = DummyFormatter()
        self.run_query_error = run_query_error
        self.ran_queries: list[tuple[str, str | None]] = []

    def run_query(self, query: str, checkpoint: str | None = None) -> None:
        if self.run_query_error is not None:
            raise self.run_query_error
        self.ran_queries.append((query, checkpoint))


def main_execute_from_cli(mycli: DummyMyCli, cli_args: DummyCliArgs) -> int:
    return execute_mode.main_execute_from_cli(cast(Any, mycli), cast(Any, cli_args))


def fake_sys(stdin_tty: bool) -> SimpleNamespace:
    return SimpleNamespace(stdin=SimpleNamespace(isatty=lambda: stdin_tty))


def test_main_execute_from_cli_returns_error_when_execute_is_missing() -> None:
    assert main_execute_from_cli(DummyMyCli(), DummyCliArgs(execute=None)) == 1


@pytest.mark.parametrize(
    ('format_name', 'original_sql', 'expected_format', 'expected_sql'),
    (
        ('csv', r'select 1\G', 'csv', 'select 1'),
        ('tsv', r'select 2\G', 'tsv', 'select 2'),
        ('table', r'select 3\G', 'ascii', 'select 3'),
        ('vertical', r'select 4\G', 'tsv', r'select 4\G'),
    ),
)
def test_main_execute_from_cli_sets_format_and_runs_query(
    monkeypatch,
    format_name: str,
    original_sql: str,
    expected_format: str,
    expected_sql: str,
) -> None:
    secho_calls: list[tuple[str, bool, str]] = []
    mycli = DummyMyCli()
    cli_args = DummyCliArgs(
        execute=original_sql,
        format=format_name,
        batch='batch.sql',
        checkpoint='cp',
    )

    monkeypatch.setattr(execute_mode, 'sys', fake_sys(stdin_tty=False))
    monkeypatch.setattr(
        execute_mode.click,
        'secho',
        lambda message, err, fg: secho_calls.append((message, err, fg)),
    )

    result = main_execute_from_cli(mycli, cli_args)

    assert result == 0
    assert mycli.main_formatter.format_name == expected_format
    assert mycli.ran_queries == [(expected_sql, 'cp')]
    assert secho_calls == [
        ('Ignoring STDIN since --execute was also given.', True, 'red'),
        ('Ignoring --batch since --execute was also given.', True, 'red'),
    ]


def test_main_execute_from_cli_does_not_warn_when_stdin_is_tty_and_batch_is_unset(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool, str]] = []
    mycli = DummyMyCli()

    monkeypatch.setattr(execute_mode, 'sys', fake_sys(stdin_tty=True))
    monkeypatch.setattr(
        execute_mode.click,
        'secho',
        lambda message, err, fg: secho_calls.append((message, err, fg)),
    )

    result = main_execute_from_cli(mycli, DummyCliArgs(execute='select 1', format='csv'))

    assert result == 0
    assert mycli.main_formatter.format_name == 'csv'
    assert mycli.ran_queries == [('select 1', None)]
    assert secho_calls == []


def test_main_execute_from_cli_reports_query_errors(monkeypatch) -> None:
    secho_calls: list[tuple[str, bool, str]] = []
    mycli = DummyMyCli(run_query_error=RuntimeError('boom'))

    monkeypatch.setattr(execute_mode, 'sys', fake_sys(stdin_tty=True))
    monkeypatch.setattr(
        execute_mode.click,
        'secho',
        lambda message, err, fg: secho_calls.append((message, err, fg)),
    )

    result = main_execute_from_cli(mycli, DummyCliArgs(execute='select 1', format='table'))

    assert result == 1
    assert mycli.main_formatter.format_name == 'ascii'
    assert mycli.ran_queries == []
    assert secho_calls == [('boom', True, 'red')]



from __future__ import annotations

from dataclasses import dataclass
from io import TextIOWrapper
import os
from pathlib import Path
import sys
from tempfile import NamedTemporaryFile
from types import SimpleNamespace
from typing import Any, Literal, cast

from click.testing import CliRunner
import pytest

import mycli.cli_runner as cli_runner
import mycli.main_modes.batch as batch_mode
import test.pytests.test_main as test_main_module
import test.utils as test_utils

noninteractive_mock_mycli = cast(Any, test_main_module).noninteractive_mock_mycli
TEMPFILE_PREFIX = cast(str, cast(Any, test_utils).TEMPFILE_PREFIX)


@dataclass
class DummyCliArgs:
    format: str = 'tsv'
    noninteractive: bool = True
    throttle: float = 0.0
    checkpoint: str | TextIOWrapper | None = None
    batch: str | None = None
    resume: bool = False


@dataclass
class DummyFormatter:
    format_name: str | None = None


class DummyLogger:
    def __init__(self) -> None:
        self.warning_messages: list[str] = []

    def warning(self, message: str) -> None:
        self.warning_messages.append(message)


class DummyMyCli:
    def __init__(self, destructive_warning: bool = False, run_query_error: Exception | None = None) -> None:
        self.main_formatter = DummyFormatter()
        self.destructive_warning = destructive_warning
        self.destructive_keywords = ('drop',)
        self.logger = DummyLogger()
        self.run_query_error = run_query_error
        self.ran_queries: list[tuple[str, str | TextIOWrapper | None, bool]] = []

    def run_query(self, query: str, checkpoint: str | TextIOWrapper | None = None, new_line: bool = True) -> None:
        if self.run_query_error is not None:
            raise self.run_query_error
        self.ran_queries.append((query, checkpoint, new_line))


class DummyFile:
    def __init__(self, name: str) -> None:
        self.name = name
        self.closed = False

    def close(self) -> None:
        self.closed = True


class DummyProgressBar:
    calls: list[list[int]] = []

    def __init__(self, *args, **kwargs) -> None:
        pass

    def __enter__(self) -> 'DummyProgressBar':
        return self

    def __exit__(self, exc_type, exc, tb) -> Literal[False]:
        return False

    def __call__(self, iterable) -> list[int]:
        values = list(iterable)
        DummyProgressBar.calls.append(values)
        return values


def dispatch_batch_statements(
    mycli: DummyMyCli,
    cli_args: DummyCliArgs,
    statements: str,
    batch_counter: int,
) -> None:
    batch_mode.dispatch_batch_statements(cast(Any, mycli), cast(Any, cli_args), statements, batch_counter)


def main_batch_with_progress_bar(mycli: DummyMyCli, cli_args: DummyCliArgs) -> int:
    return batch_mode.main_batch_with_progress_bar(cast(Any, mycli), cast(Any, cli_args))


def main_batch_without_progress_bar(mycli: DummyMyCli, cli_args: DummyCliArgs) -> int:
    return batch_mode.main_batch_without_progress_bar(cast(Any, mycli), cast(Any, cli_args))


def main_batch_from_stdin(mycli: DummyMyCli, cli_args: DummyCliArgs) -> int:
    return batch_mode.main_batch_from_stdin(cast(Any, mycli), cast(Any, cli_args))


def make_fake_sys(stdin_tty: bool, stderr_tty: bool | None = None) -> SimpleNamespace:
    stderr = SimpleNamespace(isatty=lambda: stderr_tty) if stderr_tty is not None else object()
    return SimpleNamespace(
        stdin=SimpleNamespace(isatty=lambda: stdin_tty),
        stderr=stderr,
        exit=sys.exit,
    )


def patch_progress_mode(monkeypatch, mycli_main, mycli_main_batch) -> None:
    DummyProgressBar.calls.clear()
    monkeypatch.setattr(mycli_main_batch, 'ProgressBar', DummyProgressBar)
    monkeypatch.setattr(mycli_main_batch.prompt_toolkit.output, 'create_output', lambda **kwargs: object())
    fake_sys = make_fake_sys(stdin_tty=False, stderr_tty=True)
    monkeypatch.setattr(cli_runner, 'sys', fake_sys)
    monkeypatch.setattr(mycli_main, 'sys', fake_sys)
    monkeypatch.setattr(mycli_main_batch, 'sys', fake_sys)


def invoke_click_batch(
    runner: CliRunner,
    mycli_main,
    contents: str,
    args: list[str] | None = None,
):
    with NamedTemporaryFile(prefix=TEMPFILE_PREFIX, mode='w', delete=False) as batch_file:
        batch_file.write(contents)
        batch_file.flush()

    try:
        result = runner.invoke(
            mycli_main.click_entrypoint,
            args=['--batch', batch_file.name] + (args or []),
        )
        return result, batch_file.name
    finally:
        if os.path.exists(batch_file.name):
            os.remove(batch_file.name)


def write_batch_file(tmp_path: Path, contents: str) -> str:
    batch_path = tmp_path / 'batch.sql'
    batch_path.write_text(contents, encoding='utf-8')
    return str(batch_path)


def open_checkpoint_file(tmp_path: Path, contents: str) -> TextIOWrapper:
    checkpoint_path = tmp_path / 'checkpoint.sql'
    checkpoint_path.write_text(contents, encoding='utf-8')
    return checkpoint_path.open('a', encoding='utf-8')


def test_replay_checkpoint_file_returns_zero_without_replayable_batch(tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\n')

    assert batch_mode.replay_checkpoint_file(batch_path, None, resume=True) == 0

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        with pytest.raises(batch_mode.CheckpointReplayError, match='incompatible with reading from the standard input'):
            batch_mode.replay_checkpoint_file('-', checkpoint, resume=True)


def test_replay_checkpoint_file_rejects_checkpoint_longer_than_batch(tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\n')

    with open_checkpoint_file(tmp_path, 'select 1;\nselect 2;\n') as checkpoint:
        with pytest.raises(batch_mode.CheckpointReplayError, match='Checkpoint script longer than batch script.'):
            batch_mode.replay_checkpoint_file(batch_path, checkpoint, resume=True)


@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_replay_checkpoint_file_rejects_batch_read_error(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\n')

    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', lambda _handle: (_ for _ in ()).throw(ValueError('bad batch')))

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        with pytest.raises(batch_mode.CheckpointReplayError, match=f'Error reading --batch file: {batch_path}: bad batch'):
            batch_mode.replay_checkpoint_file(batch_path, checkpoint, resume=True)


@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_replay_checkpoint_file_rejects_batch_iteration_error(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\n')

    def raise_on_next():
        raise ValueError('bad batch iterator')
        yield

    def fake_statements_from_filehandle(handle):
        if handle.name == batch_path:
            return raise_on_next()
        return iter([('select 1;', 0)])

    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', fake_statements_from_filehandle)

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        with pytest.raises(batch_mode.CheckpointReplayError, match=f'Error reading --batch file: {batch_path}: bad batch iterator'):
            batch_mode.replay_checkpoint_file(batch_path, checkpoint, resume=True)


@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_replay_checkpoint_file_rejects_checkpoint_read_error(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\n')

    def fake_statements_from_filehandle(handle):
        if handle.name == batch_path:
            return iter([('select 1;', 0)])
        return (_ for _ in ()).throw(ValueError('bad checkpoint'))

    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', fake_statements_from_filehandle)

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        with pytest.raises(batch_mode.CheckpointReplayError, match=f'Error reading --checkpoint file: {checkpoint.name}: bad checkpoint'):
            batch_mode.replay_checkpoint_file(batch_path, checkpoint, resume=True)


def test_replay_checkpoint_file_rejects_missing_files(tmp_path: Path) -> None:
    batch_path = str(tmp_path / 'missing.sql')

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        with pytest.raises(batch_mode.CheckpointReplayError, match='FileNotFoundError'):
            batch_mode.replay_checkpoint_file(batch_path, checkpoint, resume=True)


def test_replay_checkpoint_file_rejects_open_errors(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\n')

    monkeypatch.setattr(batch_mode.click, 'open_file', lambda *_args, **_kwargs: (_ for _ in ()).throw(OSError('open failed')))

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        with pytest.raises(batch_mode.CheckpointReplayError, match='OSError'):
            batch_mode.replay_checkpoint_file(batch_path, checkpoint, resume=True)


@pytest.mark.parametrize(
    ('format_name', 'batch_counter', 'expected'),
    (
        ('csv', 1, 'csv-noheader'),
        ('tsv', 1, 'tsv_noheader'),
        ('table', 1, 'ascii'),
        ('vertical', 1, 'tsv'),
        ('csv', 0, 'csv'),
        ('tsv', 0, 'tsv'),
        ('table', 0, 'ascii'),
        ('vertical', 0, 'tsv'),
    ),
)
def test_dispatch_batch_statements_sets_expected_output_format(
    format_name: str,
    batch_counter: int,
    expected: str,
) -> None:
    mycli = DummyMyCli()
    cli_args = DummyCliArgs(format=format_name, checkpoint='cp')

    dispatch_batch_statements(mycli, cli_args, 'select 1;', batch_counter)

    assert mycli.main_formatter.format_name == expected
    assert mycli.ran_queries == [('select 1;', 'cp', True)]


def test_dispatch_batch_statements_confirms_destructive_queries_before_running(monkeypatch) -> None:
    mycli = DummyMyCli(destructive_warning=True)
    cli_args = DummyCliArgs(noninteractive=False)
    opened_tty = object()

    monkeypatch.setattr(batch_mode, 'is_destructive', lambda _keywords, _statement: True)
    monkeypatch.setattr(batch_mode, 'confirm_destructive_query', lambda _keywords, _statement: True)
    monkeypatch.setattr(batch_mode, 'open', lambda _path: opened_tty, raising=False)
    monkeypatch.setattr(batch_mode, 'sys', SimpleNamespace(stdin=None))

    dispatch_batch_statements(mycli, cli_args, 'drop table demo;', 0)

    assert batch_mode.sys.stdin is opened_tty
    assert mycli.ran_queries == [('drop table demo;', None, True)]


def test_dispatch_batch_statements_skips_query_when_destructive_confirmation_is_rejected(monkeypatch) -> None:
    mycli = DummyMyCli(destructive_warning=True)
    cli_args = DummyCliArgs(noninteractive=False)

    monkeypatch.setattr(batch_mode, 'is_destructive', lambda _keywords, _statement: True)
    monkeypatch.setattr(batch_mode, 'confirm_destructive_query', lambda _keywords, _statement: False)
    monkeypatch.setattr(batch_mode, 'open', lambda _path: object(), raising=False)
    monkeypatch.setattr(batch_mode, 'sys', SimpleNamespace(stdin=None))

    dispatch_batch_statements(mycli, cli_args, 'drop table demo;', 0)

    assert mycli.ran_queries == []


def test_dispatch_batch_statements_raises_when_tty_cannot_be_opened(monkeypatch) -> None:
    mycli = DummyMyCli(destructive_warning=True)
    cli_args = DummyCliArgs(noninteractive=False)

    monkeypatch.setattr(batch_mode, 'is_destructive', lambda _keywords, _statement: True)
    monkeypatch.setattr(batch_mode, 'open', lambda _path: (_ for _ in ()).throw(OSError('tty unavailable')), raising=False)

    with pytest.raises(OSError, match='tty unavailable'):
        dispatch_batch_statements(mycli, cli_args, 'drop table demo;', 0)

    assert mycli.logger.warning_messages == ['Unable to open TTY as stdin.']


def test_dispatch_batch_statements_sleeps_and_reraises_query_errors(monkeypatch) -> None:
    mycli = DummyMyCli(run_query_error=RuntimeError('boom'))
    cli_args = DummyCliArgs(throttle=0.25)
    sleep_calls: list[float] = []
    secho_calls: list[tuple[str, bool, str]] = []

    monkeypatch.setattr(batch_mode.time, 'sleep', lambda seconds: sleep_calls.append(seconds))
    monkeypatch.setattr(
        batch_mode.click,
        'secho',
        lambda message, err, fg: secho_calls.append((message, err, fg)),
    )

    with pytest.raises(RuntimeError, match='boom'):
        dispatch_batch_statements(mycli, cli_args, 'select 1;', 1)

    assert sleep_calls == [0.25]
    assert secho_calls == []


def test_main_batch_with_progress_bar_returns_error_when_batch_is_missing() -> None:
    assert main_batch_with_progress_bar(DummyMyCli(), DummyCliArgs()) == 1


def test_main_batch_with_progress_bar_rejects_non_files(monkeypatch, tmp_path) -> None:
    messages: list[tuple[str, bool, str]] = []
    cli_args = DummyCliArgs(batch=str(tmp_path))

    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    result = main_batch_with_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert '--progress is only compatible with a plain file.' in messages[0][0]
    assert messages[0][1] is True
    assert messages[0][2] == 'red'


def test_main_batch_with_progress_bar_handles_open_errors(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []
    cli_args = DummyCliArgs(batch='missing.sql')

    monkeypatch.setattr(batch_mode.os.path, 'exists', lambda _path: False)
    monkeypatch.setattr(batch_mode.click, 'open_file', lambda _path: (_ for _ in ()).throw(FileNotFoundError()))
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    result = main_batch_with_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert messages == [('Failed to open --batch file: missing.sql', True, 'red')]


def test_main_batch_with_progress_bar_handles_counting_value_errors(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []
    count_handle = DummyFile('count')
    cli_args = DummyCliArgs(batch='statements.sql')

    monkeypatch.setattr(batch_mode.os.path, 'exists', lambda _path: False)
    monkeypatch.setattr(batch_mode.click, 'open_file', lambda _path: count_handle)
    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', lambda _handle: (_ for _ in ()).throw(ValueError('bad sql')))
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    result = main_batch_with_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert messages == [('Error reading --batch file: statements.sql: bad sql', True, 'red')]


def test_main_batch_with_progress_bar_processes_all_statements(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []
    count_handle = DummyFile('count')
    run_handle = DummyFile('run')
    open_calls: list[str] = []
    dispatch_calls: list[tuple[str, int]] = []
    cli_args = DummyCliArgs(batch='statements.sql')

    def fake_open_file(path: str) -> DummyFile:
        open_calls.append(path)
        return count_handle if len(open_calls) == 1 else run_handle

    def fake_statements_from_filehandle(handle: DummyFile):
        if handle is count_handle:
            return iter([('select 1;', 0), ('select 2;', 1)])
        return iter([('select 1;', 0), ('select 2;', 1)])

    DummyProgressBar.calls.clear()
    monkeypatch.setattr(batch_mode.os.path, 'exists', lambda _path: False)
    monkeypatch.setattr(batch_mode.click, 'open_file', fake_open_file)
    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', fake_statements_from_filehandle)
    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )
    monkeypatch.setattr(batch_mode, 'ProgressBar', DummyProgressBar)
    monkeypatch.setattr(batch_mode.prompt_toolkit.output, 'create_output', lambda **_kwargs: object())
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=False))

    result = main_batch_with_progress_bar(DummyMyCli(), cli_args)

    assert result == 0
    assert messages == [('Ignoring STDIN since --batch was also given.', True, 'yellow')]
    assert dispatch_calls == [('select 1;', 0), ('select 2;', 1)]
    assert DummyProgressBar.calls == [[0, 1]]
    assert count_handle.closed is True
    assert run_handle.closed is True


def test_main_batch_with_progress_bar_returns_error_when_dispatch_fails(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []
    count_handle = DummyFile('count')
    run_handle = DummyFile('run')
    open_calls = 0
    cli_args = DummyCliArgs(batch='statements.sql')

    def fake_open_file(_path: str) -> DummyFile:
        nonlocal open_calls
        open_calls += 1
        return count_handle if open_calls == 1 else run_handle

    def fake_statements_from_filehandle(handle: DummyFile):
        if handle is count_handle:
            return iter([('select 1;', 0)])
        return iter([('select 1;', 0)])

    monkeypatch.setattr(batch_mode.os.path, 'exists', lambda _path: False)
    monkeypatch.setattr(batch_mode.click, 'open_file', fake_open_file)
    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', fake_statements_from_filehandle)
    monkeypatch.setattr(batch_mode, 'ProgressBar', DummyProgressBar)
    monkeypatch.setattr(batch_mode.prompt_toolkit.output, 'create_output', lambda **_kwargs: object())
    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, _statement, _counter: (_ for _ in ()).throw(OSError('dispatch failed')),
    )
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    result = main_batch_with_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert messages == [('dispatch failed', True, 'red')]
    assert run_handle.closed is True


def test_main_batch_without_progress_bar_returns_error_when_batch_is_missing() -> None:
    assert main_batch_without_progress_bar(DummyMyCli(), DummyCliArgs()) == 1


def test_main_batch_without_progress_bar_handles_open_errors(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []
    cli_args = DummyCliArgs(batch='missing.sql')

    monkeypatch.setattr(batch_mode.click, 'open_file', lambda _path: (_ for _ in ()).throw(FileNotFoundError()))
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    result = main_batch_without_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert messages == [('Failed to open --batch file: missing.sql', True, 'red')]


def test_main_batch_without_progress_bar_processes_statements(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []
    batch_handle = DummyFile('run')
    dispatch_calls: list[tuple[str, int]] = []
    cli_args = DummyCliArgs(batch='statements.sql')

    monkeypatch.setattr(batch_mode.click, 'open_file', lambda _path: batch_handle)
    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', lambda _handle: iter([('select 1;', 0), ('select 2;', 1)]))
    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=False))

    result = main_batch_without_progress_bar(DummyMyCli(), cli_args)

    assert result == 0
    assert messages == [('Ignoring STDIN since --batch was also given.', True, 'red')]
    assert dispatch_calls == [('select 1;', 0), ('select 2;', 1)]
    assert batch_handle.closed is True


def test_main_batch_without_progress_bar_skips_checkpoint_prefix(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\nselect 2;\nselect 3;\n')
    dispatch_calls: list[tuple[str, int]] = []

    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    with open_checkpoint_file(tmp_path, 'select 1;\nselect 2;\n') as checkpoint:
        cli_args = DummyCliArgs(batch=batch_path, checkpoint=checkpoint, resume=True)

        result = main_batch_without_progress_bar(DummyMyCli(), cli_args)

    assert result == 0
    assert dispatch_calls == [('select 3;', 2)]


def test_main_batch_without_progress_bar_skips_only_matching_duplicate_prefix(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\nselect 1;\nselect 2;\n')
    dispatch_calls: list[tuple[str, int]] = []

    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        cli_args = DummyCliArgs(batch=batch_path, checkpoint=checkpoint, resume=True)

        result = main_batch_without_progress_bar(DummyMyCli(), cli_args)

    assert result == 0
    assert dispatch_calls == [('select 1;', 1), ('select 2;', 2)]


def test_main_batch_without_progress_bar_fails_on_mismatched_checkpoint(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\nselect 2;\n')
    dispatch_calls: list[tuple[str, int]] = []

    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    with open_checkpoint_file(tmp_path, 'select 9;\n') as checkpoint:
        cli_args = DummyCliArgs(batch=batch_path, checkpoint=checkpoint, resume=True)

        result = main_batch_without_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert dispatch_calls == []


def test_main_batch_without_progress_bar_succeeds_when_checkpoint_skips_all(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\nselect 2;\n')
    dispatch_calls: list[tuple[str, int]] = []

    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    with open_checkpoint_file(tmp_path, 'select 1;\nselect 2;\n') as checkpoint:
        cli_args = DummyCliArgs(batch=batch_path, checkpoint=checkpoint, resume=True)

        result = main_batch_without_progress_bar(DummyMyCli(), cli_args)

    assert result == 0
    assert dispatch_calls == []


def test_main_batch_with_progress_bar_skips_checkpoint_prefix_and_counts_all_statements(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\nselect 2;\nselect 3;\n')
    dispatch_calls: list[tuple[str, int]] = []

    DummyProgressBar.calls.clear()
    monkeypatch.setattr(batch_mode, 'ProgressBar', DummyProgressBar)
    monkeypatch.setattr(batch_mode.prompt_toolkit.output, 'create_output', lambda **_kwargs: object())
    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    with open_checkpoint_file(tmp_path, 'select 1;\n') as checkpoint:
        cli_args = DummyCliArgs(batch=batch_path, checkpoint=checkpoint, resume=True)

        result = main_batch_with_progress_bar(DummyMyCli(), cli_args)

    assert result == 0
    assert dispatch_calls == [('select 2;', 1), ('select 3;', 2)]
    assert DummyProgressBar.calls == [[0, 1, 2]]


def test_main_batch_with_progress_bar_returns_error_when_checkpoint_replay_fails(monkeypatch, tmp_path: Path) -> None:
    batch_path = write_batch_file(tmp_path, 'select 1;\n')
    messages: list[tuple[str, bool, str]] = []

    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    with open_checkpoint_file(tmp_path, 'select 9;\n') as checkpoint:
        cli_args = DummyCliArgs(batch=batch_path, checkpoint=checkpoint, resume=True)

        result = main_batch_with_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert messages == [(f'Error replaying --checkpoint file: {checkpoint.name}: Statement mismatch: select 9;.', True, 'red')]


def test_main_batch_without_progress_bar_returns_error_when_iteration_fails(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []
    batch_handle = DummyFile('run')
    cli_args = DummyCliArgs(batch='statements.sql')

    monkeypatch.setattr(batch_mode.click, 'open_file', lambda _path: batch_handle)
    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', lambda _handle: (_ for _ in ()).throw(ValueError('bad sql')))
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))
    monkeypatch.setattr(batch_mode, 'sys', make_fake_sys(stdin_tty=True))

    result = main_batch_without_progress_bar(DummyMyCli(), cli_args)

    assert result == 1
    assert messages == [('bad sql', True, 'red')]


def test_main_batch_from_stdin_processes_statements(monkeypatch) -> None:
    dispatch_calls: list[tuple[str, int]] = []
    batch_handle = object()

    monkeypatch.setattr(batch_mode.click, 'get_text_stream', lambda _name: batch_handle)
    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', lambda _handle: iter([('select 1;', 0), ('select 2;', 1)]))
    monkeypatch.setattr(
        batch_mode,
        'dispatch_batch_statements',
        lambda _mycli, _cli_args, statement, counter: dispatch_calls.append((statement, counter)),
    )

    result = main_batch_from_stdin(DummyMyCli(), DummyCliArgs())

    assert result == 0
    assert dispatch_calls == [('select 1;', 0), ('select 2;', 1)]


def test_main_batch_from_stdin_returns_error_for_value_errors(monkeypatch) -> None:
    messages: list[tuple[str, bool, str]] = []

    monkeypatch.setattr(batch_mode.click, 'get_text_stream', lambda _name: object())
    monkeypatch.setattr(batch_mode, 'statements_from_filehandle', lambda _handle: (_ for _ in ()).throw(ValueError('bad stdin')))
    monkeypatch.setattr(batch_mode.click, 'secho', lambda message, err, fg: messages.append((message, err, fg)))

    result = main_batch_from_stdin(DummyMyCli(), DummyCliArgs())

    assert result == 1
    assert messages == [('bad stdin', True, 'red')]


@pytest.mark.parametrize(
    ('contents', 'extra_args', 'expected_queries', 'expected_progress'),
    (
        ('select 2;', [], ['select 2;'], None),
        ('select 2; select 3;\nselect 4;\n', [], ['select 2;', 'select 3;', 'select 4;'], None),
        ('select 2;\nselect 2;\nselect 2;\n', ['--progress'], ['select 2;', 'select 2;', 'select 2;'], [[0, 1, 2]]),
        ('select 2; select 3;\nselect 4;\n', ['--progress'], ['select 2;', 'select 3;', 'select 4;'], [[0, 1, 2]]),
    ),
)
def test_click_batch_file_modes(monkeypatch, contents: str, extra_args: list[str], expected_queries: list[str], expected_progress) -> None:
    mycli_main, mycli_main_batch, MockMyCli = noninteractive_mock_mycli(monkeypatch)
    runner = CliRunner()
    MockMyCli.ran_queries = []

    if '--progress' in extra_args:
        patch_progress_mode(monkeypatch, mycli_main, mycli_main_batch)

    result, _batch_file_name = invoke_click_batch(runner, mycli_main, contents, extra_args)

    assert result.exit_code == 0
    assert MockMyCli.ran_queries == expected_queries
    if expected_progress is not None:
        assert DummyProgressBar.calls == expected_progress


def test_click_batch_file_skips_checkpoint_prefix(monkeypatch, tmp_path: Path) -> None:
    mycli_main, _mycli_main_batch, MockMyCli = noninteractive_mock_mycli(monkeypatch)
    runner = CliRunner()
    MockMyCli.ran_queries = []
    checkpoint_path = tmp_path / 'checkpoint.sql'
    checkpoint_path.write_text('select 2;\n', encoding='utf-8')

    result, _batch_file_name = invoke_click_batch(
        runner,
        mycli_main,
        'select 2;\nselect 3;\n',
        [f'--checkpoint={checkpoint_path}', '--resume'],
    )

    assert result.exit_code == 0
    assert MockMyCli.ran_queries == ['select 3;']


def test_batch_file_with_progress_requires_plain_file(monkeypatch, tmp_path) -> None:
    mycli_main, mycli_main_batch, MockMyCli = noninteractive_mock_mycli(monkeypatch)
    runner = CliRunner()

    patch_progress_mode(monkeypatch, mycli_main, mycli_main_batch)

    result = runner.invoke(
        mycli_main.click_entrypoint,
        args=['--batch', str(tmp_path), '--progress'],
    )

    assert result.exit_code != 0
    assert '--progress is only compatible with a plain file.' in result.output
    assert MockMyCli.ran_queries == []


def test_batch_file_open_error(monkeypatch) -> None:
    mycli_main, _mycli_main_batch, MockMyCli = noninteractive_mock_mycli(monkeypatch)
    runner = CliRunner()

    result = runner.invoke(mycli_main.click_entrypoint, args=['--batch', 'definitely_missing_file.sql'])

    assert result.exit_code != 0
    assert 'Failed to open --batch file' in result.output
    assert MockMyCli.ran_queries == []
